@@ -3,9 +3,9 @@ import struct
 from base64 import b16encode
 from datetime import date
 from functools import reduce
-from operator import or_
+from operator import and_, or_
 
-from django.core.exceptions import ImproperlyConfigured
+from django.core.exceptions import ImproperlyConfigured, ObjectDoesNotExist
 from django.core.files.storage import default_storage
 from django.core.files.uploadedfile import UploadedFile
 from django.core.signing import get_cookie_signer
@@ -16,6 +16,33 @@ from django.utils.timezone import datetime, now, utc
 from django.utils.translation import gettext_lazy as _
 
 
+class SimpleModelChoiceIterator(ModelChoiceIterator):
+    def __iter__(self):
+        queryset = self.queryset
+        # Can't use iterator() when queryset uses prefetch_related()
+        if not queryset._prefetch_related_lookups:
+            queryset = queryset.iterator()
+        for obj in queryset:
+            yield self.choice(obj)
+
+    def __len__(self):
+        return self.queryset.count()
+
+    def __bool__(self):
+        return self.queryset.exists()
+
+
+class GroupedModelChoiceIterator(SimpleModelChoiceIterator):
+    group_field_name = None
+
+    def choice(self, obj):
+        return (
+            ModelChoiceIteratorValue(self.field.prepare_value(obj), obj),
+            self.field.label_from_instance(obj),
+            getattr(obj, self.group_field_name),
+        )
+
+
 class IncompleteSelectMixin:
     """
     Extra interfaces for widgets not loading the complete set of choices.
@@ -24,62 +51,127 @@ class IncompleteSelectMixin:
     choices = ()
     max_prefetch_choices = 250
     search_lookup = None
+    group_field_name = None
+    filter_by = None
 
-    def __init__(self, attrs=None, choices=(), search_lookup=None):
+    def __init__(self, attrs=None, choices=(), search_lookup=None, group_field_name=None, filter_by=None):
         if search_lookup:
             self.search_lookup = search_lookup
         if isinstance(self.search_lookup, str):
             self.search_lookup = [self.search_lookup]
+        if isinstance(group_field_name, str):
+            self.group_field_name = group_field_name
+        if isinstance(filter_by, dict):
+            self.filter_by = filter_by
         super().__init__(attrs, choices)
 
-    def search(self, search_term):
+    def build_filter_query(self, filters):
+        queries = []
+        for fieldname, lookup in self.filter_by.items():
+            filtervalue = filters[fieldname]
+            if isinstance(filtervalue, list):
+                subqueries = [Q(**{lookup: val}) for val in filtervalue if val]
+                queries.append(reduce(or_, subqueries, Q()))
+            elif isinstance(filtervalue, str):
+                queries.append(Q(**{lookup: filtervalue}))
         try:
-            query = reduce(or_, (Q(**{sl: search_term}) for sl in self.search_lookup))
+            return reduce(and_, queries, Q())
+        except TypeError:
+            raise ImproperlyConfigured(f"Invalid attribute 'filter_by' in {self.__class__}.")
+
+    def build_search_query(self, search_term):
+        try:
+            return reduce(or_, (Q(**{sl: search_term}) for sl in self.search_lookup))
         except TypeError:
             raise ImproperlyConfigured(f"Invalid attribute 'search_lookup' in {self.__class__}.")
-        return self.choices.queryset.filter(query)
+
+    def format_value(self, value):
+        if value is None:
+            return []
+        if not isinstance(value, (tuple, list)):
+            value = [value]
+        return [str(v) if v is not None else "" for v in value]
+
+    def build_attrs(self, base_attrs, extra_attrs):
+        attrs = super().build_attrs(base_attrs, extra_attrs)
+        if isinstance(self.choices, SimpleModelChoiceIterator):
+            if self.choices.queryset.count() > self.max_prefetch_choices:
+                attrs['incomplete'] = True
+            if self.filter_by:
+                attrs['filter-by'] = ','.join(self.filter_by.keys())
+        return attrs
 
     def get_context(self, name, value, attrs):
         if isinstance(self.choices, ModelChoiceIterator):
-            self.optgroups = self._optgroups_model_choice
-            if self.choices.queryset.count() > self.max_prefetch_choices:
-                attrs = dict(attrs, incomplete=True)
+            if self.group_field_name:
+                self.optgroups = self._optgroups_model_choice
+                self.choices.queryset = self.choices.queryset.order_by(self.group_field_name)
+                self.choices.group_field_name = self.group_field_name
+                self.choices.__class__ = GroupedModelChoiceIterator
+            else:
+                self.optgroups = self._options_model_choice
+                self.choices.__class__ = SimpleModelChoiceIterator
         else:
             self.optgroups = self._optgroups_static_choice
         context = super().get_context(name, value, attrs)
-        options = context['widget'].pop('optgroups')
-        context['widget']['options'] = options
         return context
 
     def _optgroups_static_choice(self, name, values, attrs=None):
-        options = []
-        for val, label in self.choices:
-            val = str(val)
-            if val in values:
-                options.append({'value': val, 'label': label, 'selected': True})
-            else:
-                options.append({'value': val, 'label': label})
-        return options
+        optgroups = super().optgroups(name, values, attrs)
+        return optgroups
 
-    def _optgroups_model_choice(self, name, values, attrs=None):
-        values_set = set(values)
-        options = [{}] * len(values_set)
-        counter = 0
-        for val, label in self.choices:
+    def _options_model_choice(self, name, values, attrs=None):
+        values_list = [str(val) for val in values]
+        optgroups, counter = [], 0
+        for counter, (val, label) in enumerate(self.choices, counter):
+            if counter == self.max_prefetch_choices:
+                break
             if not isinstance(val, ModelChoiceIteratorValue):
                 continue
             val = str(val)
-            if val in values_set:
-                index = values.index(val)
-                if index < len(options):
-                    options[index] = {'value': val, 'label': label, 'selected': True}
-                values_set.remove(val)
-            elif counter < self.max_prefetch_choices:
-                options.append({'value': val, 'label': label})
-            elif not values_set:
+            if selected := val in values_list:
+                values_list.remove(val)
+            optgroups.append((None, [{'value': val, 'label': label, 'selected': selected}], counter))
+        for counter, val in enumerate(values_list, counter):
+            try:
+                obj = self.choices.queryset.get(pk=val)
+            except ObjectDoesNotExist:
+                continue
+            label = self.choices.field.label_from_instance(obj)
+            optgroups.append((None, [{'value': str(val), 'label': label, 'selected': True}], counter))
+        return optgroups
+
+    def _optgroups_model_choice(self, name, values, attrs=None):
+        values_list = [str(val) for val in values]
+        optgroups, prev_group_name, counter = [], '-', 0
+        # first handle selected values
+        for counter, val in enumerate(values_list, counter):
+            try:
+                obj = self.choices.queryset.get(pk=val)
+            except ObjectDoesNotExist:
+                continue
+            label = self.choices.field.label_from_instance(obj)
+            group_name = getattr(obj, self.group_field_name) if self.group_field_name else None
+            optgroup = list(filter(lambda item: item[0] == group_name, optgroups))
+            if optgroup:
+                optgroup[-1][1].append({'value': str(val), 'label': label, 'selected': True})
+            else:
+                subgroup = [{'value': str(val), 'label': label, 'selected': True}]
+                optgroups.append((group_name, subgroup, counter))
+        # afterwards handle the remaining values
+        for counter, (val, label, group_name) in enumerate(self.choices, counter):
+            if counter == self.max_prefetch_choices:
                 break
-            counter += 1
-        return options
+            if not isinstance(val, ModelChoiceIteratorValue):
+                continue
+            val = str(val)
+            if prev_group_name != group_name:
+                prev_group_name = group_name
+                subgroup = [{'value': val, 'label': label}]
+                optgroups.append((group_name, subgroup, counter))
+            else:
+                subgroup.append({'value': val, 'label': label})
+        return optgroups
 
 
 class Selectize(IncompleteSelectMixin, Select):
@@ -89,8 +181,8 @@ class Selectize(IncompleteSelectMixin, Select):
     template_name = 'formset/default/widgets/selectize.html'
     placeholder = _("Select")
 
-    def __init__(self, attrs=None, choices=(), search_lookup=None, placeholder=None):
-        super().__init__(attrs, choices, search_lookup)
+    def __init__(self, attrs=None, choices=(), search_lookup=None, group_field_name=None, filter_by=None, placeholder=None):
+        super().__init__(attrs, choices, search_lookup, group_field_name, filter_by)
         if placeholder is not None:
             self.placeholder = placeholder
 
@@ -101,15 +193,20 @@ class Selectize(IncompleteSelectMixin, Select):
         return attrs
 
     def _optgroups_static_choice(self, name, values, attrs=None):
-        options = [{'value': '', 'label': self.placeholder}]
+        options = [(None, [{'value': '', 'label': self.placeholder}], None)]
         options.extend(super()._optgroups_static_choice(name, values, attrs))
         return options
 
-    def _optgroups_model_choice(self, name, values, attrs=None):
-        options = [{'value': '', 'label': self.placeholder}]
-        options.extend(super()._optgroups_model_choice(name, values, attrs))
+    def _options_model_choice(self, name, values, attrs=None):
+        options = [(None, [{'value': '', 'label': self.placeholder}], None)]
+        options.extend(super()._options_model_choice(name, values, attrs))
         return options
 
+    def _optgroups_model_choice(self, name, values, attrs=None):
+        optgroups = [(None, [{'value': '', 'label': self.placeholder}], None)]
+        optgroups.extend(super()._optgroups_model_choice(name, values, attrs))
+        return optgroups
+
 
 class SelectizeMultiple(Selectize):
     allow_multiple_selected = True
